<?php

use VerifoneEcomAPI\ApiWrapper\Authentication\BasicAuth;
use VerifoneEcomAPI\ApiWrapper\Http\SimpleCurl;
use VerifoneEcomAPI\ApiWrapper\Regions\Emea;
use VerifoneEcomAPI\ApiWrapper\Regions\Na;
use VerifoneEcomAPI\ApiWrapper\Regions\NzAu;
use VerifoneEcomAPI\ApiWrapper\Settings;

if ( ! defined( 'ABSPATH' ) ) {
    exit;
} // Exit if accessed directly

// if the WC payment gateway class is not available, do nothing
if ( ! class_exists( 'WC_Payment_Gateway' ) ) {
    return;
}

/**
 * Class WC_Verifone
 */
abstract class WC_Verifone extends \WC_Payment_Gateway {

    protected $plugin_name;
    protected $version;

    /**
     * @var BasicAuth
     */
    protected $verifone_auth_helper;

    /**
     * @var Settings
     */
    protected $verifone_settings;

    /**
     * @var WC_Logger
     */
    protected $logger;

    /**
     * @var Emea|Na|NzAu
     */
    protected $api_url_mapper;

    /**
     * @var SimpleCurl
     */
    protected $simple_curl;

    const PLUGIN_ID = 'verifone_hosted_cart';

    /**
     * WC_Verifone_Hosted_Cart constructor.
     */
    public function __construct() {
        //Initialize all the basics components of the plugin
        load_plugin_textdomain( 'verifone-hosted', false, 'verifone-hosted/languages' );
        $this->plugin_name                  = 'Verifone Hosted Cart Payment';
        $this->method_description           = __( 'Card payments and alternative payment methods powered by Verifone', 'verifone-hosted' );
        $this->method_long_description      = __( 'Accept card payments and other payment options available in your Verifone account. Hosted cart or iFrame experiences available.', 'verifone-hosted' );
        $this->method_description_link_text = __( 'Read the documentation.', 'verifone-hosted' );
        $this->method_description_link_url  = 'https://verifone.cloud/docs/online-payments/plugins/woocommerce';
        $this->supports                     = [
            'refunds',
            'subscriptions',
            'subscription_cancellation',
            'subscription_suspension',
            'subscription_reactivation',
            'subscription_amount_changes',
            'subscription_date_changes',
            'subscription_payment_method_change_customer',
        ];
        $this->method_title                 = __( 'Verifone Hosted Cart', 'verifone-hosted' );
        $this->version                      = '1.0.1';
        $this->id                           = self::PLUGIN_ID;
        $this->enabled                      = $this->get_option( 'enabled' );
        $this->has_fields                   = false;
        // Define user set variables
        $this->title       = $this->get_option( 'title' );
        $this->description = $this->get_option( 'description', __( 'Payments powered by', 'verifone-hosted' ) . ' <a href="https://www.verifone.com/" target="_blank">Verifone</a>' );

        // Load the settings
        $this->init_form_fields();
        $this->init_settings();

        $payment_button_description = $this->get_option( 'checkout_button_text' );
        if ( empty( $payment_button_description ) ) {
            $payment_button_description = 'Proceed to Verifone Hosted Cart';
        }
        $this->order_button_text = __( $payment_button_description, 'verifone-hosted' );

        $settings = new Settings();

        $settings
            ->setSendTaxRates( $this->trueOrFalse( 'send_tax_rates' ) )
            ->setIsInvoice4UEnabled( $this->trueOrFalse( 'invoice4u_enabled' ) )
            ->setInvoice4UReceiptType( $this->get_option( 'invoice4u_receipt_type' ) )
            ->setTest( $this->trueOrFalse( 'test' ) )
            ->setRegion( $this->get_option( 'region' ) )
            ->setApiUserId( $this->get_option( 'user_id' ) )
            ->setApiKey( $this->get_option( 'api_key' ) )
            ->set3dsEnable( $this->trueOrFalse( '_3ds_enable' ) )
            ->setOverlayOnPlaceOrder( $this->trueOrFalse( 'overlay_on_place_order' ) )
            ->setEmptyOnPlaceOrder( $this->trueOrFalse( 'empty_on_place_order' ) )
            ->setManageOrders( $this->trueOrFalse( 'manage_orders' ) )
            ->set3dsContractId( $this->get_option( '_3ds_contract_id' ) )
            ->setIsCardEnabled( $this->trueOrFalse( 'card_enabled' ) )
            ->setPaymentContractId( $this->get_option( 'payment_contract_id' ) )
            ->setEntityId( $this->get_option( 'entity_id' ) )
            ->setCurlVerifySsl( $this->trueOrFalse( 'curl_verify_ssl' ) )
            ->setTransactionType( $this->get_option( 'transaction_type' ) )
            ->setThemeId( $this->get_option( 'theme_id' ) )
            ->setCompleteOrderOnPayment( $this->trueOrFalse( 'complete_order_on_payment' ) )
            ->setPaypalContractId( $this->get_option( 'paypal_contract_id' ) )
            ->setIsApplePayEnabled( $this->trueOrFalse( 'apple_pay_enabled' ) )
            ->setApplePayScaComplianceLevel( $this->get_option( 'apple_pay_sca_compliance_level' ) )
            ->setApplePayContractId( $this->get_option( 'apple_pay_contract_id' ) )
            ->setIsGooglePayEnabled( $this->trueOrFalse( 'google_pay_enabled' ) )
            ->setGooglePayScaComplianceLevel( $this->get_option( 'google_pay_sca_compliance_level' ) )
            ->setGooglePayContractId( $this->get_option( 'google_pay_contract_id' ) )
            ->setIsPaypalEnabled( $this->trueOrFalse( 'paypal_enabled' ) )
            ->setIsSwishPayEnabled( $this->trueOrFalse( 'swish_pay_enabled' ) )
            ->setIsAbsInstallmentEnabled( $this->trueOrFalse( 'abs_installment_enabled' ) )
            ->setIsKlarnaEnabled( $this->trueOrFalse( 'klarna_enabled' ) )
            ->setIsGpp2Enabled( $this->trueOrFalse( 'gpp2_enabled' ) )
            ->setGpp2ContractId( $this->get_option( 'gpp2_contract_id' ) )
            ->setIsSubscriptionEnabled( $this->trueOrFalse( 'subscription_enabled' ) )
            ->setSubscriptionId( $this->get_option( 'subscription_id' ) )
            ->setIsCardFraudCheckEnabled( $this->trueOrFalse( 'fraud_check_enabled' ) )
            ->setFraudCheckContractId( $this->get_option( 'fraud_check_contract_id' ) )
            ->setIsAbsInstallmentEnabled( $this->trueOrFalse( 'abs_installment_enabled' ) )
            ->setAbsInstallmentThreshholds( $this->get_option( 'abs_installment_threshholds' ) )
            ->setIsMobilePayEnabled( $this->trueOrFalse( 'mobile_pay_enabled' ) )
            ->setMobilePayScaComplianceLevel( $this->get_option( 'mobile_pay_sca_compliance_level' ) )
            ->setMobilePayContractId( $this->get_option( 'mobile_pay_contract_id' ) )
            ->setIsVippsEnabled( $this->trueOrFalse( 'vipps_enabled' ) )
            ->setVippsScaComplianceLevel( $this->get_option( 'vipps_sca_compliance_level' ) )
            ->setVippsContractId( $this->get_option( 'vipps_contract_id' ) )
            ->setOoverrideRedirectUrl( $this->get_option( 'override_redirect_url' ) )
            ->setPaymentType(
                Settings::PAYMENT_TYPE_HPP == $this->get_option( 'payment_type' )
                    ? Settings::PAYMENT_TYPE_HPP
                    : Settings::PAYMENT_TYPE_IFRAME
            );

        $this->verifone_settings = $settings;
        $this->simple_curl       = new SimpleCurl();

        switch ( $this->get_option( 'region' ) ) {
            case 'NZAU':
                $this->api_url_mapper = new NzAu( $this->verifone_settings );
                break;
            case 'NA':
                $this->api_url_mapper = new Na( $this->verifone_settings );
                break;
            case 'EMEA':
                $this->api_url_mapper = new Emea( $this->verifone_settings );
                break;
            default:
                // will happen only on the first installation
                $this->api_url_mapper = new NzAu( $this->verifone_settings );
        }

        $this->logger = new WC_Logger();

        $this->verifone_auth_helper = new BasicAuth( $this->verifone_settings );
    }

    public function get_webhook_response_url() {
        return WC()->api_request_url( 'woo-verifone-webhook' );
    }

    public function api_response($message = null, $statusCode = 200) {
        status_header($statusCode);

        if (is_array($message)) {
            header("Content-Type: application/json");
            echo wp_json_encode($message);
        } else
            echo $message;

        exit;
    }

    public function api_response_ok($message = null) {
        $this->api_response(['info' => $message], 200);
    }

    public function api_response_error($message = null, $code = null) {
        $this->api_response(['error' => $message, 'code' => $code], 200);
    }

    public function trueOrFalse( $optionName ) {
        return 'yes' == strtolower( $this->get_option( $optionName ) );
    }

    /** Check if order has only downloadable products
     *
     * @param WC_Order $order
     *
     * @return bool
     */
    public function isOrderOnlyVirtual( WC_Order $order ) {
        foreach ( $order->get_items() as $item ) {
            $product = wc_get_product( $item->get_product_id() );
            if ( ! $product->is_virtual() ) {
                return false;
            }
        }

        return true;
    }
}
