<?php

use VerifoneEcomAPI\ApiWrapper\Http\ApiClient\Client;
use VerifoneEcomAPI\ApiWrapper\Http\SimpleCurl;
use VerifoneEcomAPI\ApiWrapper\Settings;

/**
 * Class WC_Verifone_Hosted_Gateway
 */
class WC_Verifone_Hosted_Gateway extends WC_Verifone {

    private $client;
    private $refundStatusOk = [
        VERIFONE_TRANSACTION_AUTHORIZED,
        VERIFONE_TRANSACTION_AUTHORISED,
        VERIFONE_TRANSACTION_INITIATED,
        VERIFONE_TRANSACTION_REFUNDED,
        VERIFONE_TRANSACTION_REFUND_SETTLEMENT_COMPLETED
    ];
    const SUPPORTED_I18N_LANGUAGES = [ 'en', 'da', 'fi', 'fr', 'he', 'no', 'pl', 'pt', 'sv' ];
    const NORWAY_CODES = [ 'nb', 'nn' ];

    /**
     * @inheritDoc
     */
    public function __construct() {
        parent::__construct();
        $this->add_actions();

        $this->client = new Client( $this->api_url_mapper, $this->verifone_settings, $this->verifone_auth_helper, new SimpleCurl() );
    }

    /**
     * Adds various actions to be used later on
     * Mostly hooks
     */
    private function add_actions() {
        add_action( 'woocommerce_update_options_payment_gateways_' . $this->id, [ $this, 'process_admin_options', ] );
        add_action( 'woocommerce_receipt_' . $this->id, [ $this, 'receipt_page' ] );
        add_action( 'wp_enqueue_scripts', [ $this, 'enqueue_script' ] );
        add_action( 'woocommerce_update_options_payment_gateways_' . $this->id, [ $this, 'reset_token' ] );
        add_action( 'woocommerce_api_payment_response', [ $this, 'verifone_check_payment_response' ] );
        add_action( 'woocommerce_api_manual_capture_payment', [ $this, 'manual_capture_payment' ] );
        add_action( 'woocommerce_order_status_cancelled', [ $this, 'cancel_payment' ] );
        add_action( 'woocommerce_order_status_completed', [ $this, 'complete_payment' ] );
        add_action( 'woocommerce_order_status_refunded', [ $this, 'cancel_payment' ] );
        add_action( 'woocommerce_order_item_add_action_buttons', [ $this, 'verifone_checkout_capture' ], 10, 2 );
        add_action( 'admin_enqueue_scripts', array( $this, 'admin_scripts' ) );
        add_action( 'woocommerce_api_woo-verifone-webhook', [ $this, 'process_webhook' ] );
        add_action( 'woocommerce_get_sections_checkout', [ $this, 'add_subsection' ] );
        add_action( 'woocommerce_order_item_add_action_buttons', [ $this, 'add_void_transaction_button' ], 10, 1 );
        add_action( 'woocommerce_order_item_add_action_buttons', [ $this, 'hide_order_refund_button' ] );
        add_action( 'woocommerce_api_cancel_payment', [ $this, 'manual_cancel_payment' ] );
        add_action( 'woocommerce_scheduled_subscription_payment_' . $this->id, [ $this, 'scheduled_subscription_payment' ], 10, 2 );

        if ( $this->verifone_settings->getPaymentType() === Settings::PAYMENT_TYPE_IFRAME ) {
            add_action( 'woocommerce_api_verifone_iframe_handle_payment_request', [ $this, 'handle_payment_request' ] );
        }
    }

    /**
     * Triggered when a new payment is done for a subscription
     *
     * @param $renewal_total
     * @param WC_Order $renewal_order
     *
     * @return bool
     * @throws Exception
     */
    public function scheduled_subscription_payment( $renewal_total, WC_Order $renewal_order ) {

        $subscriptionId = null;

        foreach ( $renewal_order->get_meta_data() as $val ) {
            $arr = $val->get_data();
            if ( $arr['key'] === '_subscription_renewal' ) {
                $subscriptionId = $arr['value'];
                break;
            }
        }
        $reference = get_post_meta( $subscriptionId, VERIFONE_ORDER_STORED_CREDENTIAL_REFERENCE )[0];

        if ( ! $reference ) {
            $renewal_order->update_status( 'failed', 'Unable to charge card, subscription reference not found' );
            $this->logger->debug( 'The subscription reference for order ' . $renewal_order->get_id() . ' could not be found!' );
            throw new Exception( 'The subscription reference for order ' . $renewal_order->get_id() . ' could not be found!' );
        }

        $payload = [
            'amount'                    => absint( wc_format_decimal( ( $renewal_total * 100 ), wc_get_price_decimals() ) ),
            'currency_code'             => get_option( 'woocommerce_currency' ),
            'merchant_reference'        => (string) $renewal_order->get_id(),
            'payment_provider_contract' => $this->verifone_settings->getPaymentContractId(),
            'dynamic_descriptor'        => (string) $renewal_order->get_id(),
            'stored_credential'         => [
                'stored_credential_type' => 'CHARGE',
                'processing_model'       => 'RECURRING',
                'reference'              => $reference
            ]
        ];
        try {
            $result = $this->client->postSubscription( $payload );
            $renewal_order->add_order_note( __( 'Transaction ID: ' . $result['id'] ) );
            if ( isset( $result['status'] ) && in_array( $result['status'], [ 'AUTHORIZED', 'SETTLEMENT_COMPLETED' ] ) ) {
                WC_Subscriptions_Manager::process_subscription_payments_on_order( $renewal_order );
                $renewal_order->update_status( 'completed', 'Payment success for subscription ID: ' . $subscriptionId );
            } else {
                WC_Subscriptions_Manager::process_subscription_payment_failure_on_order( $renewal_order );
                $renewal_order->update_status( 'failed', 'Payment failed for subscription ID: ' . $subscriptionId );
            }
        } catch ( \VerifoneEcomAPI\ApiWrapper\Http\HttpException $exception ) {
            // update the status to failed
            $renewal_order->update_status( 'failed', 'Payment failed for subscription ID: ' . $subscriptionId );
            $this->logger->debug( sprintf( 'There was an error in the api response, last known error was "%s", in file %s, at line %s. Error message: %s',
                json_last_error(), __FILE__, __LINE__, $exception->getMessage() ) );

            return false;
        }

        return true;
    }

    public function verifone_checkout_capture( \WC_Order $order ) {
        $authType      = 'FINAL_AUTH'; //default value
        $manualCapture = 'NO';
        foreach ( $order->get_meta_data() as $val ) {
            $arr = $val->get_data();
            if ( $arr['key'] === '__verifone_order_transaction_type' ) {
                $authType = $arr['value'];
            }
            if ( $arr['key'] === '__verifone_order_capture_manual' ) {
                $manualCapture = $arr['value'];
            }
        }

        //display buttons only when the order is in PENDING and the MANUAL CAPTURE is set (PRE or FINAL auth)
        if ( $manualCapture === 'yes' && in_array( $order->get_status(), [ 'pending', 'processing' ] ) ) {
            $url = get_site_url() . '/?wc-api=manual_capture_payment&order_id=' . $order->get_id();
            include( plugin_dir_path( __FILE__ ) . '../templates/html-order-capture.php' );
        }

        add_action( 'woocommerce_order_item_add_action_buttons', array( $this, 'add_capture_button' ) );

        add_action( 'wp_ajax_wc_' . $this->id . '_capture_charge', array( $this, 'ajax_process_capture' ) );
    }

    /**
     *
     * @throws Exception
     */
    public function manual_capture_payment() {
        ob_start();
        global $wpdb;

        $params   = $_POST;
        $order_id = $params['order_id'];
        $order    = wc_get_order( $order_id );
        $reason   = $params['reason'];
        $amount   = (float) $params['amount'];

        if ( ! $order ) {
            wp_send_json( [ 'result' => 'failure', 'messages' => 'Missing or invalid order ID.' ], 402 );
        }
        $total = $order->get_total();
        if ( ! current_user_can( 'edit_shop_order', $order_id ) ) {
            wp_send_json( [ 'result' => 'failure', 'messages' => 'Unauthorized' ], 401 );
        }
        if ( $amount > $total ) {
            wp_send_json( [ 'result' => 'failure', 'messages' => 'Forbidden - capture amount cannot be greater than the order total!' ], 403 );
        }

        if ( $amount ) { //full or partial capture
            $this->voidOrCapture( $order_id, 'capture', absint( wc_format_decimal( ( $amount * 100 ), wc_get_price_decimals() ) ) );
        } else {  // full capture only
            $amount = $total;
            $this->voidOrCapture( $order_id );
        }
        $note = 'Order amount captured: ' . $amount . ' ' . $order->get_currency() . '. Refund is not available within the first 24-48 hours.';
        if ( $reason ) {
            $note .= ' *Reason: ' . $reason;
        }
        $order->update_status( 'completed', $note );

        //update order's total with the paid amount
        $data  = [ 'meta_value' => $amount ];
        $where = [ 'meta_key' => '_order_total', 'post_id' => $order_id ];
        $wpdb->update( $wpdb->prefix . 'postmeta', $data, $where );

        wp_send_json( [ 'result' => 'success', 'message' => $note ], 200 );
    }


    /**
     * @param $order_id
     */
    public function cancel_payment( $order_id ) {
        $this->voidOrCapture( $order_id, 'void' );
    }

    /**
     * @param $order_id
     */
    public function complete_payment( $order_id ) {
        $this->voidOrCapture( $order_id );
    }

    /**
     * ajax callable wrapper for process_payment
     *
     * @access public
     * @return array
     */
    public function handle_payment_request() {
        ob_start();
        if ( ! empty( $_POST['terms-field'] ) && empty( $_POST['terms'] ) ) {
            wp_send_json( [
                'result'   => 'failure',
                'messages' => 'Please read and accept the terms and conditions to proceed with your order.',
                'step'     => 'existing_order',
            ] );
        }

        if ( $_SERVER['REQUEST_METHOD'] !== 'POST' || ! isset( $_POST['woocommerce-pay-nonce'] ) || ! isset( $_POST['order_id'] ) ) {
            return;
        }

        $nonce_value = $_POST['woocommerce-pay-nonce'];

        if ( ! wp_verify_nonce( $nonce_value, 'woocommerce-pay' ) ) {
            return;
        }

        $response = $this->process_payment( $_POST['order_id'] );

        wp_send_json( $response );
    }

    /**
     * @param $order_id
     * @param string $type
     * @param null $amount
     *
     * @return true|void
     */
    private function voidOrCapture( $order_id, string $type = 'capture', $amount = null ) {

        // we make this check if the order is updated by changing the status and if 'manageOrders' is set to TRUE,
        // if we press the 'capture amount' button we don't make the check
        $calledBy = debug_backtrace()[1]['function'];
        if ( in_array( $calledBy, [ 'complete_payment', 'cancel_payment' ] ) && ! $this->verifone_settings->getManageOrders() ) {
            return true;
        }
        $order          = wc_get_order( $order_id );
        $payment_method = $order->get_meta( VERIFONE_PAYMENT_METHOD_USED );

        if ( self::PLUGIN_ID === $order->get_payment_method() ) {
            if ( ! $order->get_meta( VERIFONE_CAPTURE_MANUAL ) ) {
                return;
            }

            if ( $order->get_meta( VERIFONE_CAPTURE_MANUAL ) && $order->get_meta( VERIFONE_CAPTURE_MANUAL ) !== 'yes' ) {
                return;
            }
            if ( $order->get_meta( VERIFONE_CAPTURE_DONE ) && $order->get_meta( VERIFONE_CAPTURE_DONE ) === 'yes' ) {
                return;
            }
            try {
                $transaction_id = $order->get_meta( VERIFONE_ORDER_TRANSACTION_ID );
                if ( ! $amount ) {
                    $amount = absint( wc_format_decimal( ( $order->get_total() * 100 ), wc_get_price_decimals() ) );
                }
                $payload = [ 'amount' => $amount ];

                if ( $type == 'void' ) {
                    $result = $this->client->postVoidAuthorization( $payload, $transaction_id );
                }
                if ( $type == 'capture' ) {
                    if ( VERIFONE_PAYMENT_METHOD_PAYPAL === $payment_method ) {
                        $payload['amount'] = [
                            'value'        => $payload['amount'],
                            'currencyCode' => $order->get_currency(),
                        ];
                        $result            = $this->client->postPaypalCapture( $payload, $transaction_id );
                    } else {
                        $result = $this->client->postCapture( $payload, $transaction_id );
                    }
                }

                // a 400 request means the capture was already done
                // thus we can ignore it
                if ( isset( $result['code'] ) && $result['code'] == 400 ) {
                    return;
                }

                if ( isset( $result['status'] ) && $result['status'] == VERIFONE_ENV_CAPTURE_EVENT_TYPE_STATUS_AUTHORIZED ) {
                    $message = sprintf( 'Order with transaction ID %s was authorized. Authorization ID %s', $transaction_id, $result['id'] );
                }

                if ( isset( $result['status'] ) && $result['status'] == VERIFONE_ENV_CAPTURE_EVENT_TYPE_STATUS_SETTLEMENT_COMPLETED ) {
                    $message = sprintf( 'Order with transaction ID %s was settled. Authorization ID %s', $transaction_id, $result['id'] );
                }

                if ( isset( $result['status'] ) && $result['status'] == VERIFONE_ENV_CAPTURE_EVENT_TYPE_STATUS_VOIDED ) {
                    $message = sprintf( 'Order with transaction ID %s was voided. Void ID %s', $transaction_id, $result['id'] );
                }

                if ( isset( $message ) ) {
                    $order->add_order_note( $message, false );
                    $order->update_meta_data( '__verifone_order_capture_done', 'yes' );
                } else {
                    $order->add_order_note( 'No change detected for order, please try setting the status again', false );
                }
                if ( is_callable( [ $order, 'save' ] ) ) {
                    $order->save();
                }
            } catch ( \VerifoneEcomAPI\ApiWrapper\Http\HttpException $exception ) {
                // update the status to failed
                $order->update_status( 'failed', 'Unable to change order status, please try again later' );
                $this->logger->debug( sprintf( 'There was an error in the api response, last known error was "%s", in file %s, at line %s. Error message: %s',
                    json_last_error(), __FILE__, __LINE__, $exception->getMessage() ) );
            }
        }
    }

    /**
     * Resets the token on ANY setting change
     */
    public function reset_token() {
        $token = get_option( '__verifone_api_token' );
        if ( false !== $token ) {
            delete_option( '__verifone_api_token' );
        }
    }

    /**
     * Includes a javascript file to be used in the frontend
     */
    public function enqueue_script() {
        if ( $this->verifone_settings->getOverlayOnPlaceOrder() ) {
            wp_enqueue_script( 'verifone_hosted_cart_script', '/wp-content/plugins/verifone-hosted/assets/js/verifone_hosted_cart.js' );
        }

        if ( $this->verifone_settings->getPaymentType() === Settings::PAYMENT_TYPE_IFRAME ) {
            wp_enqueue_script( 'verifone_hosted_iframe_script', '/wp-content/plugins/verifone-hosted/assets/js/verifone_hosted_cart_iframe.js', [], false, true );
        }
    }

    /**
     * @inheritDoc
     */
    public function admin_options() {
        require_once plugin_dir_path( __FILE__ ) . '../templates/admin-options.php';
    }

    /**
     * Includes a javascript file to be used in the backend
     */
    public function admin_scripts() {
        wp_enqueue_script( 'verifone_admin_script', '/wp-content/plugins/verifone-hosted/assets/js/verifone_admin.js' );
    }

    /**
     * @inheritDoc
     */
    public function init_form_fields() {
        require_once plugin_dir_path( __FILE__ ) . '../templates/init_form_fields.php';
        $current_subsection = ( isset( $_GET['vhc-tab'] ) && ! empty( $_GET['vhc-tab'] ) ) ? $_GET['vhc-tab'] : '';
        switch ( $current_subsection ) {
            case 'verifone-account-settings-tab':
            default:
                $this->form_fields = getVerifoneHostedCartVerifoneAccountSettingsFields();
                break;
            case 'card-payment-settings-tab':
                $this->form_fields = getVerifoneHostedCartCardPaymentSettingsFields();
                break;
            case 'alternative-payment-options-tab':
                $this->form_fields = getVerifoneHostedCartAlternativePaymentOptionsFields();
                break;
            case 'display-settings-tab':
                $this->form_fields = getVerifoneHostedCartDisplaySettingsFields();
                break;
            case 'advanced-settings-tab':
                $this->form_fields = getVerifoneHostedCartAdvancedSettingsFields();
                break;
        }

    }

    /**
     * @param int $order_id
     * @param null $amount
     * @param string $reason
     *
     * @return bool|void|WP_Error
     */
    public function process_refund( $order_id, $amount = null, $reason = '' ) {
        $order = wc_get_order( $order_id );

        if ( $order->get_payment_method() !== self::PLUGIN_ID ) {
            return;
        }

        // if the order was not captured and try to refund, we void the transaction
        if ( $order->get_status() === 'pending' &&
             in_array( $order->get_meta( VERIFONE_ORDER_TRANSACTION_TYPE ), [ 'FINAL_AUTH', 'PRE_AUTH' ] ) ) {

            $this->cancel_payment( $order->get_id() );

            return true;
        }

        $transaction_id = $order->get_meta( VERIFONE_ORDER_TRANSACTION_ID );
        $payment_method = $order->get_meta( VERIFONE_PAYMENT_METHOD_USED );
        if ( ! $transaction_id ) {
            $this->logger->debug( sprintf( 'Tried to refund order with order ID %s but it has no registered transaction ID, aborting.', $order_id ) );

            return new WP_Error( 'verifone_refund_error', __( 'Refund Error: Unable to refund transaction.' ) );
        }
        $date = new DateTime( 'now' );

        $amount = (float) $amount;
        $amount = $amount * 100;

        if ( VERIFONE_PAYMENT_METHOD_PAYPAL === $payment_method ) {
            $refund = [
                'amount' => [
                    'value'        => (int) $amount,
                    'currencyCode' => $order->get_currency(),
                ]
            ];
        } else {
            $refund = [
                'amount'            => (int) $amount,
                'id'                => $order->get_meta( VERIFONE_ORDER_TRANSACTION_ID ),
                'reference_id'      => $order->get_id(),
                'created_date_time' => "{$date->format( 'Y-m-d\TH:i:s\Z' )}",
            ];
        }

        if ( '' !== $reason ) {
            $refund['reason'] = $reason;
        }
        try {
            $response = ( VERIFONE_PAYMENT_METHOD_PAYPAL === $payment_method )
                ? $this->client->postPaypalRefund( $refund, $transaction_id )
                : $this->client->postRefund( $refund, $transaction_id );
        } catch ( Exception $exception ) {
            $this->logger->error( 'There was an error in the API response: ' . PHP_EOL .
                                  'Sent request: ' . json_encode( $refund ) . PHP_EOL .
                                  'Received response: ' . $exception->getMessage() . PHP_EOL .
                                  'To endpoint: ' . $this->api_url_mapper->getRefundUrl( $transaction_id )
            );

            return new WP_Error( 'verifone_refund_error', __LINE__ . ': ' . __( 'Refund Error: Unable to refund transaction.' ) );
        }

        if ( ! isset( $response['id'] )
             || ! isset( $response['status'] )
             || ! in_array( $response['status'], $this->refundStatusOk )
        ) {
            return new WP_Error( 'verifone_refund_error', __LINE__ . ': ' . __( 'Refund Error: Unable to refund transaction.' ) );
        }

        $order->update_meta_data( VERIFONE_ORDER_REFUND_ID, $response['id'] );

        if ( 'INITIATED' === $response['status'] ) {
            $order->add_order_note( __( sprintf( 'Refunded %s out of a total of %s was initiated, pending processing', $amount / 100, $order->get_total() ) ) );
        } else {
            $order->add_order_note( __( sprintf( 'Refunded %s out of a total of %s from order', $amount / 100, $order->get_total() ) ) );
        }

        $order->save_meta_data();
        $order->save();

        return true;
    }

    /**
     * @param int $order_id
     *
     * @return array|\WP_Error
     * @throws Exception
     */
    public function process_payment( $order_id ) {
        if ( $this->verifone_settings->is3dsEnabled() && null === $this->verifone_settings->get3dsContractId() ) {
            $this->logger->debug( '3DS is enabled but no 3DS contract ID is provided in the configuration. Please fill in the field "3DS Contract Id" or disable the field "Enable/Disable 3d-secure" in the plugin settings.' );

            return $this->failure();
        }

        $order = wc_get_order( $order_id );
        if ( ! $order instanceof WC_Order ) {
            return $this->failure();
        }

        $payment_gateways = WC()->payment_gateways->payment_gateways();
        if ( ! isset( $payment_gateways[ self::PLUGIN_ID ] ) ) {
            $this->logger->debug( sprintf( 'Attempted to set payment method as %s however said payment method does not exist. This error should not happen unless the payment method is misspelled.',
                self::PLUGIN_ID ) );

            return $this->failure();
        }

        $order->set_payment_method( $payment_gateways[ self::PLUGIN_ID ] );
        $order->save();

        global $woocommerce;

        $locale = $this->s40( get_locale(), 2 );

        if ( in_array( $locale, self::NORWAY_CODES ) ) {
            $locale = "no";
        }

        //create the customer - will return NULL if the customer could not be created
        $customer = $this->getCustomer( $order );
        if ( ! $customer ) {
            return $this->failure();
        }

        $transaction_mode = wp_is_mobile() ? 'P' : 'S';
        $orderAmount      = absint( wc_format_decimal( ( $order->get_total() * 100 ), wc_get_price_decimals() ) );
        $order->update_meta_data( VERIFONE_CAPTURE_MANUAL, $this->verifone_settings->isManualCapture() ? 'yes' : 'no' );
        $order->update_meta_data( VERIFONE_ORDER_TRANSACTION_TYPE, $this->verifone_settings->getAuthType() );
        $order->save_meta_data();
        $redirectUrl = add_query_arg( [ 'wc-api' => 'payment_response', 'pm' => 'verifone_hosted', ], home_url( '/' ) );

        if ( ! empty ( $this->verifone_settings->getOverrideRedirectUrl() ) ) {
            $redirectUrl = $this->verifone_settings->getOverrideRedirectUrl();
        }
        $checkout = [
            'entity_id'          => $this->verifone_settings->getEntityId(),
            'currency_code'      => get_option( 'woocommerce_currency' ),
            'amount'             => $orderAmount,
            'return_url'         => $redirectUrl,
            'customer'           => $customer,
            'merchant_reference' => (string) $order->get_id(),
            'configurations'     => [],
            'i18n'               => [
                'default_language' => in_array( $locale, self::SUPPORTED_I18N_LANGUAGES ) ? $locale : 'en'
            ],
        ];

        $this->card( $checkout, $transaction_mode );
        $this->mobilPay( $checkout, $order, $transaction_mode );
        $this->paypal( $checkout );
        $this->klarna( $checkout, $orderAmount );
        $this->googlePay( $checkout, $transaction_mode );
        $this->swish( $checkout, $transaction_mode );
        $this->applePay( $checkout, $transaction_mode );
        $this->vipps( $checkout, $order, $transaction_mode );
        $this->gpp2( $checkout );
        $this->subscriptions( $checkout, $order );
        $this->installments( $checkout, $order );
        $this->invoice4u( $checkout, $order );

        if ( ! empty ( $this->verifone_settings->getThemeId() ) ) {
            $checkout['theme_id'] = $this->verifone_settings->getThemeId();
        }

        try {
            $result = $this->client->postCheckout( $checkout );

        } catch ( Exception $exception ) {
            $this->logger->debug(
                'There was an error in the API response: ' . PHP_EOL .
                'Sent request: ' . json_encode( $checkout ) . PHP_EOL .
                'Received response: ' . $exception->getMessage() . PHP_EOL .
                'To endpoint: ' . $this->api_url_mapper->postCheckoutUrl()
            );

            return $this->failure();
        }

        if ( ! isset( $result['url'] ) ) {
            $this->logger->debug( sprintf( 'The "url" key is not present in the verifone response, unable to continue, in file %s, at line %s. API response %s',
                __FILE__, __LINE__, json_encode( $result ) ) );

            return $this->failure();
        }

        if ( $this->verifone_settings->getPaymentType() === Settings::PAYMENT_TYPE_IFRAME ) {
            // the host seems to be different from any url we currently use
            $scheme = parse_url( $result['url'], PHP_URL_SCHEME );
            $host   = parse_url( $result['url'], PHP_URL_HOST );

            return [
                'result'  => 'success',
                'payload' => wp_json_encode( [
                    'url' => sprintf( '%s://%s/v2/loader.js?checkoutId=%s', $scheme, $host, $result['id'] ),
                ] ),
            ];
        }

        if ( $this->verifone_settings->getEmptyOnPlaceOrder() ) {
            $woocommerce->cart->empty_cart();
        }

        return [
            'result'   => 'success',
            'redirect' => $result['url'],
        ];
    }

    /**
     * @throws Exception
     */
    public function verifone_check_payment_response() {
        $params = $_GET;

        if ( ! isset( $params['pm'] ) || $params['pm'] != 'verifone_hosted' ) {
            return;
        }

        if ( ! isset( $params['checkout_id'] ) ) {
            wc_add_notice( __( 'There was an unexpected error during the processing of your payment. Please contact the store for an up-to-date status.',
                'woocommerce-gateway-verifone-hosted-checkout' ), 'error' );
            $this->logger->debug( sprintf( 'The checkout_id parameter could not be found, in order to proceed it must exist. Current GET params: %s',
                json_encode( $_GET ) ) );
            wp_safe_redirect( wc_get_page_permalink( 'cart' ) );
            exit;
        }

        try {
            $response = $this->client->getCheckout( $params['checkout_id'] );
        } catch ( Exception $exception ) {
            $this->logger->debug( 'There was an error in the API response: ' . PHP_EOL .
                                  'GET parameters: ' . json_encode( $params ) . PHP_EOL .
                                  'Received response: ' . $exception->getMessage() . PHP_EOL .
                                  'To endpoint: ' . $this->api_url_mapper->getCheckoutUrl( $params['checkout_id'] )
            );

            wc_add_notice( __( 'There was an unexpected error during the processing of your payment. Please contact the store for an up-to-date status.',
                'woocommerce-gateway-verifone-hosted-checkout' ), 'error' );
            wp_safe_redirect( wc_get_page_permalink( 'cart' ) );
            exit;
        }

        if ( ! isset( $response['events'] )
             || ! isset( $response['merchant_reference'] )
             || ! isset( $response['transaction_id'] )
        ) {
            $this->logger->debug( sprintf( 'The events or merchant_reference or transactionId parameter could not be found in response, in order to proceed it must exist. API response %s',
                json_encode( $response ) ) );
            wc_add_notice( __( 'There was an unexpected error during the processing of your payment. Please contact the store for an up-to-date status.',
                'woocommerce-gateway-verifone-hosted-checkout' ), 'error' );
            wp_safe_redirect( wc_get_page_permalink( 'cart' ) );
            exit;
        }

        $order = wc_get_order( (int) $response['merchant_reference'] );
        if ( ! $order instanceof WC_Order ) {
            $this->logger->debug( 'There was a request for an order that doesn\'t exist in current shop! Requested params: ' . json_encode( $_GET ) );
            wc_add_notice( __( 'There was an unexpected error during the processing of your payment. Please contact the store for an up-to-date status.',
                'woocommerce-gateway-verifone-hosted-checkout' ), 'error' );
            wp_safe_redirect( wc_get_page_permalink( 'cart' ) );
            exit;
        }

        // save the transaction id too
        $order->update_meta_data( VERIFONE_ORDER_TRANSACTION_ID, $response['transaction_id'] );
        $order->update_meta_data( VERIFONE_PAYMENT_METHOD_USED, $response['payment_method_used'] );
        $order->save_meta_data();

        // set the order status based on the checkout response
        // I was only able to get 3 statuses to get pushed back
        foreach ( $response['events'] as $event ) {
            if ( VERIFONE_ENV_CHECKOUT_EVENT_TYPE_TRANSACTION_SUCCESS == $event['type'] ) {
                $order->add_order_note( __( 'Transaction ID: ' . $response['transaction_id'] ) );

                if ( ! $event['details']['details']['auto_capture'] ) {
                    $order->update_status( 'processing' );
                } elseif ( $order->get_meta( VERIFONE_CAPTURE_MANUAL ) === 'yes' ) {
                    if ( 'pending' != $order->get_status() ) {
                        $order->add_order_note( __( 'The payment is not done yet (transaction is in PRE/FINAL AUTH status), use the CAPTURE button to fully/partial capture the amount. 
                    Attempting to refund the order in part or in full, before capturing, will void the authorization and cancel the payment.
                    Order status changed to Pending.' ), false );
                    }
                    $order->update_status( 'processing' );
                } else {
                    $order->payment_complete();
                    if ( $this->verifone_settings->getCompleteOrderOnPayment() || $this->isOrderOnlyVirtual( $order ) ) {

                        //Cannot update status to completed for already active subscriptions.
                        if ( ! ( $order instanceof WC_Subscription ) || 'active' !== $order->get_status() ) {
                            $order->update_status( 'completed' );
                        }
                    }
                    $order->add_order_note( __( "Order payment is completed." ), false );
                }
                $this->hasSubscriptions( $event, $order );
            }

            if ( VERIFONE_ENV_CHECKOUT_EVENT_TYPE_TRANSACTION_FAILED == $event['type'] ) {
                $order->update_status( 'failed' );
                $order->add_order_note( __( "Order status changed to failed" ), false );
            }

            if ( VERIFONE_ENV_CHECKOUT_EVENT_TYPE_TIMEOUT_ERROR == $event['type'] ) {
                $order->update_status( 'failed' );
                $order->add_order_note( __( "Order status changed to failed" ), false );
            }

            if ( VERIFONE_ENV_CHECKOUT_EVENT_TYPE_TRANSACTION_DECLINED == $event['type'] ) {
                $order->update_status( 'failed' );
                $order->add_order_note( __( "Payment was declined" ), false );
            }
        }

        $order->save();

        $redirect_url = $order->get_checkout_order_received_url();
        if ( wp_redirect( $redirect_url ) ) {
            exit;
        }
    }

    private function hasSubscriptions( array $event, WC_Order $order ) {
        if ( isset( $event['details'] ) &&
             isset( $event['details']['stored_credential'] ) &&
             isset( $event['details']['stored_credential']['reference'] ) &&
             function_exists( 'wcs_get_subscriptions_for_order' ) ) {

            if ( $order instanceof WC_Subscription ) {
                $subscriptions = [ $order ];
            } else {
                $subscriptions = wcs_get_subscriptions_for_order( $order->get_id() );
            }

            foreach ( $subscriptions as $subscription ) {
                $subscription->add_meta_data( VERIFONE_ORDER_STORED_CREDENTIAL_REFERENCE, $event['details']['stored_credential']['reference'], true );
                $subscription->save();
            }

            $order->add_order_note( __( "Subscription reference: " . $event['details']['stored_credential']['reference'] ) );
        }
    }

    public function process_webhook() {
        $input = file_get_contents( 'php://input' );
        if ( ! $input ) {
            $this->api_response_error( 'Unable to process data' );
        }

        $data = json_decode( $input, true );
        if ( json_last_error() !== JSON_ERROR_NONE ) {
            $this->logger->error( __CLASS__ . '@' . __FUNCTION__ . ' line ' . __LINE__ . ': Unable to decode response from API' );
            $this->api_response_error( 'Unable to process data', __LINE__ );
        }

        if ( $this->stringStartsWith( $data['eventType'], 'Checkout' ) ) {
            try {
                $checkout         = $this->client->getCheckout( $data['recordId'] );
                $data['recordId'] = $checkout['transaction_id'];
            } catch ( Exception $exception ) {
                $this->logger->error( __CLASS__ . '@' . __FUNCTION__ . ' line ' . __LINE__ . ': There was a request for an order: ' . $exception->getMessage() );
                $this->api_response_error( 'Unable to process data', __LINE__ );
            }
        }

        try {
            $response = $this->client->getTransaction( $data['recordId'] );
        } catch ( Exception $exception ) {
            $this->logger->error( __CLASS__ . '@' . __FUNCTION__ . ' line ' . __LINE__ . ': There was a request for an order: ' . $exception->getMessage() );
            $this->api_response_error( 'Unable to retrieve transaction', __LINE__ );
        }

        $order = null;
        if ( ! empty( $response['merchant_reference'] ) ) {
            $order = wc_get_order( (int) $response['merchant_reference'] );
            if ( ! $order instanceof WC_Order ) {
                $this->logger->debug( __CLASS__ . '@' . __FUNCTION__ . ' line ' . __LINE__ . '. There was a request for an order (' . $response['merchant_reference'] . ') that doesn\'t exist in current shop! Record ID: ' . $data['recordId'] . ' , response: ' . json_encode( $response ) );
                $this->api_response_error( 'Unable to process data', __LINE__ );
            }
            $order->update_meta_data( VERIFONE_PAYMENT_METHOD_USED, $response['payment_product'] );
            $order->update_meta_data( VERIFONE_ORDER_TRANSACTION_ID, $data['recordId'] );
        } elseif ( ! empty( $response['id'] ) ) {
            $order = wc_get_orders( [
                'limit'        => 1,
                'orderby'      => 'date',
                'order'        => 'DESC',
                'meta_key'     => VERIFONE_ORDER_REFUND_ID,
                'meta_value'   => $response['id'],
                'meta_compare' => '=',
            ] );
            if ( $order ) {
                $order = $order[0];
            }
        }
        if ( ! $order instanceof WC_Order ) {
            $this->logger->debug( __CLASS__ . '@' . __FUNCTION__ . ' line ' . __LINE__ . '. There was a request for the Record ID: ' . $data['recordId'] . ' , response: ' . json_encode( $response ) );
            $this->api_response_error( 'Unable to process data', __LINE__ );
        }

        $this->logOrderComment( $data['eventType'], $order );
        switch ( $response['status'] ) {
            case VERIFONE_TRANSACTION_CAPTURE_AUTHORISED:
            case VERIFONE_TRANSACTION_CAPTURE_SETTLED:
            case VERIFONE_TRANSACTION_AUTHORISED:
            case VERIFONE_TRANSACTION_SALE_AUTHORISED:
            case VERIFONE_TRANSACTION_SALE_SETTLED:
            case VERIFONE_TRANSACTION_CARD_VERIFICATION_AUTHORISED:
                $order->payment_complete();
                if ( $this->verifone_settings->getCompleteOrderOnPayment() || $this->isOrderOnlyVirtual( $order ) ) {
                    $order->update_status( 'completed' );
                }
                $order->add_order_note( __( "Order payment is completed." ), false );
                $this->hasSubscriptions( $data, $order );
                break;

            case VERIFONE_TRANSACTION_AUTHORISATION_AUTHORISED:
                if ( 'pending' == $order->get_status() ) {
                    $order->add_order_note( __( "Order payment is authorised. Order status changed to processing." ), false );
                }
                $order->update_status( 'processing' );
                break;

            case VERIFONE_TRANSACTION_PENDING:
            case VERIFONE_TRANSACTION_PREAUTH_AUTHORISED:
                if ( 'pending' != $order->get_status() ) {
                    $order->add_order_note( __( 'The payment is not done yet (transaction is in PRE/FINAL AUTH status), use the CAPTURE button to fully/partial capture the amount. 
                    Attempting to refund the order in part or in full, before capturing, will void the authorization and cancel the payment.
                    Order status changed to Pending.' ), false );
                }
                $order->update_status( 'processing' );
                break;

            case VERIFONE_TRANSACTION_REFUND_AUTHORISED:
            case VERIFONE_TRANSACTION_REFUND_REFUNDED:
            case VERIFONE_TRANSACTION_REFUNDED:
            case VERIFONE_TRANSACTION_REFUND_SETTLED:
                $order->add_order_note( __( "Order payment is refunded." ), false );
                break;

            default:
                if ( VERIFONE_WEBHOOK_EVENT_TYPE_REFUND_DECLINED === $data['eventType'] ) {
                    $verifoneResponseAmount = (float) $response['amount'];

                    $refunds = $order->get_refunds();
                    foreach ( $refunds as $refund ) {
                        $refundAmount = (float) $refund->get_amount();
                        if ( 'completed' === $refund->get_status() && $verifoneResponseAmount === $refundAmount ) {
                            $refund->delete();
                            break;
                        }
                    }
                } else {
                    $order->update_status( 'failed' );
                    $order->add_order_note( __( "Order status changed to failed" ), false );
                }
        }

        $this->api_response_ok();
    }

    /**
     * @return array
     */
    private function failure() {
        return [
            'result'   => 'failure',
            'messages' => 'There has been an error processing your order',
        ];
    }

    public function post_capture( $order ) {

    }

    /**
     * substrings an string
     *
     * @param $string
     * @param int $length
     *
     * @return false|string
     */
    private function s40( $string, $length = 40 ) {

        return substr( $string, 0, $length );
    }

    private function sanitizeValue( &$value, $replace = ',' ) {
        $charset = [ '\\', '/', ':', ';', '\'', '-' ];
        $value   = str_replace( $charset, $replace, $value );
    }

    private function sanitizeCustomerInfo( &$customerInfo ) {
        foreach ( array_keys( $customerInfo ) as $key ) {
            $replace = ',';

            if ( in_array( $key, [ 'phone', 'postal_code' ] ) ) {
                $replace = '';
            }

            $this->sanitizeValue( $customerInfo[ $key ], $replace );
        }
    }

    /**
     * Creates the customer object returns the customer_id
     *
     * @param $order
     *
     * @return mixed|null
     */
    private function getCustomer( $order ) {
        try {

            $customer_info = [];
            $phone         = preg_replace( '/[^0-9]+/', '', $this->s40( $order->get_billing_phone(), 20 ) );

            $customer_info['entity_id'] = $this->verifone_settings->getEntityId();

            if ( ! empty( $phone ) ) {
                $customer_info['phone_number']     = $phone;
                $customer_info['billing']['phone'] = $phone;
            }
            //billing
            if ( ! empty( $order->get_billing_company() ) ) {
                $customer_info['company_name'] = $this->s40( $order->get_billing_company(), 100 );
            }
            if ( ! empty( $order->get_billing_email() ) ) {
                $customer_info['email_address'] = $this->s40( $order->get_billing_email(), 255 );
            }
            if ( ! empty( $order->get_billing_address_1() ) ) {
                $customer_info['billing']['address_1'] = $this->s40( $order->get_billing_address_1() );
            }
            if ( ! empty( $order->get_billing_city() ) ) {
                $customer_info['billing']['city'] = $this->s40( $order->get_billing_city() );
            }
            if ( ! empty( $order->get_billing_country() ) ) {
                $customer_info['billing']['country_code'] = strtoupper( $order->get_billing_country() );
            }
            if ( ! empty( $order->get_billing_first_name() ) ) {
                $customer_info['billing']['first_name'] = $this->s40( $order->get_billing_first_name(), 50 );
            }
            if ( ! empty( $order->get_billing_last_name() ) ) {
                $customer_info['billing']['last_name'] = $this->s40( $order->get_billing_last_name(), 50 );
            }
            if ( ! empty( $order->get_billing_postcode() ) ) {
                $customer_info['billing']['postal_code'] = $this->s40( $order->get_billing_postcode(), 10 );
            }
            if ( ! empty( $order->get_billing_state() ) ) {
                $customer_info['billing']['state'] = $this->s40( $order->get_billing_state(), 50 );
            }
            if ( ! empty( $order->get_billing_address_2() ) ) {
                $customer_info['billing']['address_2'] = $this->s40( $order->get_billing_address_2() );
            }

            //shipping
            if ( '' === $order->get_shipping_address_1() ) {
                $customer_info['shipping'] = $customer_info['billing'];
            } else {
                if ( ! empty( $phone ) ) {
                    $customer_info['shipping']['phone'] = $phone;
                }
                if ( ! empty( $order->get_shipping_address_1() ) ) {
                    $customer_info['shipping']['address_1'] = $this->s40( $order->get_shipping_address_1() );
                }
                if ( ! empty( $order->get_shipping_city() ) ) {
                    $customer_info['shipping']['city'] = $this->s40( $order->get_shipping_city() );
                }
                if ( ! empty( $order->get_shipping_country() ) ) {
                    $customer_info['shipping']['country_code'] = strtoupper( $order->get_shipping_country() );
                }
                if ( ! empty( $order->get_shipping_first_name() ) ) {
                    $customer_info['shipping']['first_name'] = $this->s40( $order->get_shipping_first_name(), 50 );
                }
                if ( ! empty( $order->get_shipping_last_name() ) ) {
                    $customer_info['shipping']['last_name'] = $this->s40( $order->get_shipping_last_name(), 50 );
                }
                if ( ! empty( $order->get_shipping_postcode() ) ) {
                    $customer_info['shipping']['postal_code'] = $this->s40( $order->get_shipping_postcode(), 10 );
                }
                if ( ! empty( $order->get_shipping_state() ) ) {
                    $customer_info['shipping']['state'] = $this->s40( $order->get_shipping_state(), 50 );
                }
                if ( ! empty( $order->get_shipping_address_2() ) ) {
                    $customer_info['shipping']['address_2'] = $this->s40( $order->get_shipping_address_2() );
                }
            }

            if ( ! empty( $customer_info['phone_number'] ) ) {
                $this->sanitizeValue( $customer_info['phone_number'], '' );
            }

            if ( ! empty( $customer_info['email_address'] ) ) {
                $this->sanitizeValue( $customer_info['email_address'], '' );
            }

            if ( ! empty( $customer_info['company_name'] ) ) {
                $this->sanitizeValue( $customer_info['company_name'], ',' );
            }

            $this->sanitizeCustomerInfo( $customer_info['billing'] );
            $this->sanitizeCustomerInfo( $customer_info['shipping'] );

            $result = $this->client->postCustomer( $customer_info );

            return $result['id'];
        } catch ( \VerifoneEcomAPI\ApiWrapper\Http\HttpException $exception ) {
            $this->logger->debug( 'There was an error in the API response: ' . PHP_EOL .
                                  'Sent request: ' . json_encode( $customer_info ) . PHP_EOL .
                                  'Received response: ' . $exception->getMessage() . PHP_EOL .
                                  'To endpoint: ' . $this->api_url_mapper->postCustomerUrl()
            );

            return null;
        }
    }

    /**
     * @param $checkout
     * @param $orderAmount
     *
     * @return void
     */
    private function klarna( &$checkout, $orderAmount ) {
        if ( $this->verifone_settings->getIsKlarnaEnabled() ) {
            $checkout['configurations']['klarna'] = [
                'capture_now' => ! $this->verifone_settings->isManualCapture(),
            ];
            // we don't need the line items if we have subscription
            if ( ! class_exists( 'WC_Subscriptions_Cart' ) ||
                 ( class_exists( 'WC_Subscriptions_Cart' ) &&
                   ! WC_Subscriptions_Cart::cart_contains_subscription() ) ) {
                $checkout['line_items'] [] = [
                    'name'             => substr( preg_replace( '/[^A-Za-z0-9\-]/', ' ', get_bloginfo( 'name' ) ), 0, 255 ),
                    "quantity"         => 1,
                    "tax_rate"         => 0,
                    "total_tax_amount" => 0,
                    "unit_price"       => $orderAmount,
                    "total_amount"     => $orderAmount,
                ];
            }
        }
    }

    /**
     * @param $checkout
     * @param $transaction_mode
     *
     * @return void
     */
    private function applePay( &$checkout, $transaction_mode ) {
        if ( $this->verifone_settings->isApplePayEnabled() ) {
            $checkout['configurations']['apple_pay'] = [
                'card'        => [
                    'sca_compliance_level' => $this->verifone_settings->getApplePayScaComplianceLevel(),
                    'payment_contract_id'  => $this->verifone_settings->getApplePayContractId(),
                ],
                'capture_now' => ! $this->verifone_settings->isManualCapture(),
            ];

            if ( $this->verifone_settings->getApplePayScaComplianceLevel() === 'FORCE_3DS' ) {
                $checkout['configurations']['apple_pay']['card']['threed_secure'] = [
                    'threeds_contract_id' => $this->verifone_settings->get3dsContractId(),
                    'transaction_mode'    => $transaction_mode,
                ];
            }
        }
    }

    /**
     * @param $checkout
     * @param $transaction_mode
     *
     * @return void
     */
    private function googlePay( &$checkout, $transaction_mode ) {
        if ( $this->verifone_settings->isGooglePayEnabled() ) {

            $checkout['configurations']['google_pay'] = [
                'card'        => [
                    'sca_compliance_level' => $this->verifone_settings->getGooglePayScaComplianceLevel(),
                    'payment_contract_id'  => $this->verifone_settings->getGooglePayContractId(),
                ],
                'capture_now' => ! $this->verifone_settings->isManualCapture(),
            ];
            if ( in_array( $this->verifone_settings->getGooglePayScaComplianceLevel(), [ 'WALLET', 'FORCE_3DS' ] ) ) {
                $checkout['configurations']['google_pay']['card']['threed_secure'] = [
                    'threeds_contract_id' => $this->verifone_settings->get3dsContractId(),
                    'transaction_mode'    => $transaction_mode,
                ];
            }
        }
    }

    /**
     * @param $checkout
     * @param $transaction_mode
     *
     * @return void
     */
    private function swish( &$checkout, $transaction_mode ) {
        if ( $this->verifone_settings->isSwishPayEnabled() ) {
            $checkout['configurations']['swish'] = [
                'shopper_interaction' => $transaction_mode === "P" ? 'MCOMMERCE' : 'ECOMMERCE',
            ];
        }
    }

    /**
     * @param $checkout
     *
     * @return void
     */
    private function paypal( &$checkout ) {
        if ( $this->verifone_settings->isPaypalEnabled() ) {
            $checkout['configurations']['paypal'] = [
                'capture_now'         => ! $this->verifone_settings->isManualCapture(),
                'payment_contract_id' => $this->verifone_settings->getPaypalContractId(),
            ];
        }
    }

    private function invoice4u( &$checkout, $order ) {
        if ( $this->verifone_settings->getIsInvoice4UEnabled() ) {
            $checkout['receipt_type'] = $this->verifone_settings->getInvoice4UReceiptType();
            // we add the line item products

            $all_tax_rates = [];
            $tax_classes   = WC_Tax::get_tax_classes(); // Retrieve all tax classes.
            if ( ! in_array( '', $tax_classes ) ) { // Make sure "Standard rate" (empty class name) is present.
                array_unshift( $tax_classes, '' );
            }
            foreach ( $tax_classes as $tax_class ) { // For each tax class, get all rates.
                $taxes         = WC_Tax::get_rates_for_tax_class( $tax_class );
                $all_tax_rates = array_merge( $all_tax_rates, $taxes );

            }
            foreach ( $order->get_items() as $k => $itemData ) {
                $item     = $itemData->get_data();
                $lineItem = [
                    'name'         => substr( preg_replace( '/[^A-Za-z0-9\-]/', ' ', $item['name'] ), 0, 248 ),
                    "quantity"     => $item['quantity'],
                    "unit_price"   => absint( wc_format_decimal( ( $order->get_item_total( $itemData, true, true ) * 100 ), wc_get_price_decimals() ) ),
                    "total_amount" => absint( wc_format_decimal( ( $order->get_line_total( $itemData, true, true ) * 100 ), wc_get_price_decimals() ) ),
                ];
                if ( $this->verifone_settings->getSendTaxRates() ) {
                    foreach ( $all_tax_rates as $rate ) {
                        if ( array_key_first( $item['taxes']['total'] ) == $rate->tax_rate_id ) {
                            $lineItem['tax_rate'] = (int) $rate->tax_rate;
                        }
                    }
                }

                $checkout['line_items'] [] = $lineItem;
            }
            $shippingRatePercent = 0;
            foreach ( WC()->session->get( 'shipping_for_package_0' )['rates'] as $method_id => $shippingRate ) {
                if ( WC()->session->get( 'chosen_shipping_methods' )[0] == $method_id ) {
                    foreach ( $all_tax_rates as $rate ) {
                        if ( array_key_first( $shippingRate->get_taxes() ) == $rate->tax_rate_id ) {
                            $shippingRatePercent = (int) $rate->tax_rate;
                            break;
                        }
                    }
                    break;
                }
            }
            // add shipping cost if exists
            if ( $order->get_data()['shipping_total'] > 0 ) {
                $totalAmount = absint( wc_format_decimal( ( $order->get_data()['shipping_total'] + $order->get_data()['shipping_tax'] ) * 100, wc_get_price_decimals() ) );

                $checkout['line_items'][] = [
                    'name'         => __( 'Shipping fee' ),
                    "quantity"     => 1,
                    "unit_price"   => $totalAmount,
                    "tax_rate"     => $shippingRatePercent,
                    "total_amount" => $totalAmount
                ];
            }
            // add discount if exists
            // we don't use it at this time
//            if ( $order->get_data()['discount_total'] > 0 ) {
//                $discount = absint( wc_format_decimal( ( ($order->get_data()['discount_total'] + $order->get_data()['discount_tax'] ) * 100 ), wc_get_price_decimals() ) );
//                $checkout['line_items'][] = [
//                    'name'         => __( 'Discount' ),
//                    "quantity"     => 1,
//                    "unit_price"   => - $discount,
//                    "total_amount" => - $discount,
//                    "tax_rate"     => 0,
//                ];
//            }
        }
    }


    /**
     * @param $checkout
     *
     * @return void
     */
    private function gpp2( &$checkout ) {
        if ( $this->verifone_settings->getIsGpp2Enabled() ) {
            $checkout['configurations']['gpp2'] = [
                'payment_contract_id' => $this->verifone_settings->getGpp2ContractId(),
            ];
        }
    }

    /**
     * @param $checkout
     * @param $order
     *
     * @return void
     */
    private function installments( &$checkout, $order ) {
        if ( $this->verifone_settings->getIsCardEnabled() && $this->verifone_settings->isAbsInstallmentEnabled() ) {
            $installment = $this->verifone_settings->getAbsInstallmentThreshholdForAmount( $order->get_total() );
            if ( $installment && $installment >= Settings::MIN_INSTALLMENT_PAYMENT_DISPLAY_NUMBER ) {
                $checkout['configurations']['card']['credit_term'] = 'INSTALMENT_STANDARD';
                $checkout['configurations']['card']['instalment']  = [
                    'max_number_of_instalments' => $installment
                ];
            } else {
                $checkout['configurations']['card']['credit_term'] = 'STANDARD';
            }
        }

    }

    /**
     * @param $checkout
     * @param $transaction_mode
     * @param $order
     * @param string $type
     *
     * @return void
     */
    private function card( &$checkout, $transaction_mode ) {
        if ( $this->verifone_settings->getIsCardEnabled() ) {
            $threeDS = [];

            $settings = [
                'capture_now'         => ! $this->verifone_settings->isManualCapture(),
                'payment_contract_id' => $this->verifone_settings->getPaymentContractId(),
                'authorization_type'  => $this->verifone_settings->getAuthType(),
            ];

            if ( $this->verifone_settings->getIsCardFraudCheckEnabled() ) {
                $settings['fraud_protection_contract'] = $this->verifone_settings->getFraudCheckContractId();
            }

            if ( $this->verifone_settings->is3dsEnabled() ) {
                $threeDS = [
                    'threed_secure' => [
                        'threeds_contract_id' => $this->verifone_settings->get3dsContractId(),
                        'enabled'             => $this->verifone_settings->is3dsEnabled(),
                        'transaction_mode'    => $transaction_mode,
                    ]
                ];
            }

            $checkout['configurations']['card'] = array_merge( $settings, $threeDS );
        }
    }

    private function subscriptions( &$checkout, $order, $type = 'SIGNUP' ) {
        if ( $this->verifone_settings->getIsCardEnabled() ) {
            //check if order has any subscription && the subscription is enabled
            if ( class_exists( 'WC_Subscriptions_Cart' ) &&
                 ( WC_Subscriptions_Cart::cart_contains_subscription() || $order instanceof WC_Subscription ) &&
                 $this->verifone_settings->getIsSubscriptionEnabled() ) {

                // unset all other payment method except CARD
                // subscriptions are supported only by CARD
                foreach ( $checkout['configurations'] as $key => $value ) {
                    if ( $key !== 'card' ) {
                        unset( $checkout['configurations'][ $key ] );
                    }
                }

                $checkout['configurations']['card']['token_preference'] = [
                    'token_scope' => $this->verifone_settings->getSubscriptionId(),
                    'token_type'  => 'REUSE'
                ];

                $checkout['configurations']['card']['stored_credential'] = [
                    'stored_credential_type'   => $type,
                    // "first_payment" --- Set to true if this signup request also is the first payment. Most relevant for signup using a minor amount,
                    // to indicate if that minor amount is just a signup account verification amount or an actual first payment amount as well.
                    // Default on this endpoint is that zero amount is not a payment,
                    // but any amount provided is a payment. Use this indicator to explicitly tell the intention
                    'first_payment'            => $order->get_total() > 0,
                    'processing_model_details' => [
                        'processing_model'  => 'RECURRING',
                        'payment_frequency' => [
                            'time_unit' => $this->getOrderBillingPeriodForSubscription( $order ),
                            'value'     => (string) absint( wc_format_decimal( ( WC_Subscriptions_Order::get_recurring_total( $order ) * 100 ), wc_get_price_decimals() ) )
                        ]
                    ],
                ];
            }
        }
    }

    /**
     * @param $order
     *
     * @return string
     */
    private function getOrderBillingPeriodForSubscription( $order ) {
        if ( $order instanceof WC_Subscription ) {
            $billing_period = $order->get_billing_period();
        } else {
            foreach ( wcs_get_subscriptions_for_order( $order, array( 'order_type' => 'parent' ) ) as $subscription ) {

                // Find the billing period discount for all recurring items
                if ( empty( $product_id ) ) {
                    $billing_period = $subscription->get_billing_period();
                    break;
                } else {
                    // We want the billing period for a specific item (so we need to find if this subscription contains that item)
                    foreach ( $subscription->get_items() as $line_item ) {
                        if ( wcs_get_canonical_product_id( $line_item ) == $product_id ) {
                            $billing_period = $subscription->get_billing_period();
                            break 2;
                        }
                    }
                }
            }
        }

        return strtoupper( $billing_period );
    }

    /**
     * @param $checkout
     * @param $order
     * @param $transaction_mode
     *
     * @return void
     */
    private function mobilPay( &$checkout, $order, $transaction_mode ) {
        if ( $this->verifone_settings->isMobilePayEnabled() ) {
            $checkout['configurations']['mobile_pay'] = [
                "dynamic_descriptor"  => (string) $order->get_id(),
                'capture_now'         => ! $this->verifone_settings->isManualCapture(),
                'payment_contract_id' => $this->verifone_settings->getMobilePayContractId(),
                'card'                => [
                    "sca_compliance_level" => $this->verifone_settings->getMobilePayScaComplianceLevel(),
                    "authorization_type"   => $this->verifone_settings->getAuthType(),
                ]
            ];
            if ( $this->verifone_settings->is3dsEnabled() ) {
                $checkout['configurations']['mobile_pay']['card']['threed_secure'] = [
                    "threeds_contract_id" => $this->verifone_settings->get3dsContractId(),
                    "transaction_mode"    => $transaction_mode,
                    "enabled"             => true,
                ];
            }
        }
    }

    /**
     * @param $checkout
     * @param $order
     * @param $transaction_mode
     *
     * @return void
     */
    private function vipps( &$checkout, $order, $transaction_mode ) {
        if ( $this->verifone_settings->isVippsEnabled() ) {
            $checkout['configurations']['vipps'] = [
                'dynamic_descriptor'  => (string) $order->get_id(),
                'capture_now'         => ! $this->verifone_settings->isManualCapture(),
                'payment_contract_id' => $this->verifone_settings->getVippsContractId(),
                'card'                => [
                    'sca_compliance_level' => $this->verifone_settings->getVippsScaComplianceLevel(),
                    'authorization_type'   => $this->verifone_settings->getAuthType()
                ]
            ];

            if ( 'NONE' !== $this->verifone_settings->getVippsScaComplianceLevel() && $this->verifone_settings->get3dsContractId() ) {
                $checkout['configurations']['vipps']['card']['threed_secure'] = [
                    'enabled'             => true,
                    'threeds_contract_id' => $this->verifone_settings->get3dsContractId(),
                    'total_items'         => str_pad( count( $order->get_items() ), 2, '0', STR_PAD_LEFT ),
                    'transaction_mode'    => $transaction_mode,
                ];
            }
        }
    }

    public function add_subsection( $settings_tabs ) {
        $current_page = ( isset( $_GET['tab'] ) && ! empty( $_GET['tab'] ) ) ? $_GET['tab'] : '';
        if ( $current_page == 'checkout' && $settings_tabs !== null ) {
            $current_section = ( isset( $_GET['section'] ) && ! empty( $_GET['section'] ) ) ? $_GET['section'] : '';
            if ( $current_section == 'verifone_hosted_cart' ) {
                $sections = array(
                    'verifone-account-settings-tab'   => __( 'Verifone account settings', 'woocommerce-verifone-account-settings-tab' ),
                    'display-settings-tab'            => __( 'Display settings', 'woocommerce-display-settings-tab' ),
                    'card-payment-settings-tab'       => __( 'Card payment settings', 'woocommerce-card-payment-settings-tab' ),
                    'alternative-payment-options-tab' => __( 'Alternative payment options', 'woocommerce-alternative-payment-options-tab' ),
                    'advanced-settings-tab'           => __( 'Advanced settings', 'woocommerce-advanced-settings-tab' ),
                );
                echo '<h2 style="display: inline-block">' . $this->method_title . '</h2>';
                wc_back_link( __( 'Return to payments', 'woocommerce' ), admin_url( 'admin.php?page=wc-settings&tab=checkout' ) );
                echo '<p>' . $this->method_long_description . ' <a href="' . $this->method_description_link_url . '" target="_blank">' . $this->method_description_link_text . '</a></p>';
                echo '<p>' . __( 'Register this link to your Verifone Account', 'verifone-hosted' ) . '</p>
                    <div style="display: inline-block; padding: 5px 20px; border: 1px solid #BBBBBB; margin-bottom: 30px; border-radius: 2px; background-color: #EEEEEE;">
                        <button style="margin-right: 10px; cursor: pointer; display: inline-block; border: 0" id="webhook_copy_button" type="button">
                            <svg xmlns="http://www.w3.org/2000/svg" width="16" height="16" fill="currentColor" class="bi bi-clipboard" viewBox="0 0 16 16">
                              <path d="M4 1.5H3a2 2 0 0 0-2 2V14a2 2 0 0 0 2 2h10a2 2 0 0 0 2-2V3.5a2 2 0 0 0-2-2h-1v1h1a1 1 0 0 1 1 1V14a1 1 0 0 1-1 1H3a1 1 0 0 1-1-1V3.5a1 1 0 0 1 1-1h1v-1z"/>
                              <path d="M9.5 1a.5.5 0 0 1 .5.5v1a.5.5 0 0 1-.5.5h-3a.5.5 0 0 1-.5-.5v-1a.5.5 0 0 1 .5-.5h3zm-3-1A1.5 1.5 0 0 0 5 1.5v1A1.5 1.5 0 0 0 6.5 4h3A1.5 1.5 0 0 0 11 2.5v-1A1.5 1.5 0 0 0 9.5 0h-3z"/>
                            </svg>
                        </button>
                    <p style="display: inline-block" id="webhook_response_url">' . $this->get_webhook_response_url() . '</p>
                    </div>';
                echo '<br/>';
                echo '<ul style="margin-bottom: 20px" class="subsubsub">';
                $array_keys         = array_keys( $sections );
                $key                = "vhc-tab";
                $current_subsection = ( isset( $_GET['vhc-tab'] ) && ! empty( $_GET['vhc-tab'] ) ) ? $_GET['vhc-tab'] : 'default';
                foreach ( $sections as $id => $label ) {
                    $url = admin_url( 'admin.php?page=wc-settings&tab=checkout&section=verifone_hosted_cart&' . $key . '=' . $id );
                    echo '<li><a href="' . esc_url( $url ) . '" class="' . ( $current_subsection === $id || ( $current_subsection === 'default' && $id === 'verifone-account-settings-tab' ) ? 'current' : '' ) . '">' . esc_html( $label ) . '</a> ' . ( end( $array_keys ) === $id ? '' : '|' ) . ' </li>';
                }
                echo '</ul><br class="clear" />';
            }
        }
    }

    public function generate_radio_html( $key, $data ) {
        $field_key = $this->get_field_key( $key );
        $defaults  = array(
            'title'             => '',
            'disabled'          => false,
            'class'             => '',
            'css'               => '',
            'placeholder'       => '',
            'type'              => 'text',
            'desc_tip'          => false,
            'description'       => '',
            'custom_attributes' => array(),
        );

        $data = wp_parse_args( $data, $defaults );

        $defaultValue = null;
        foreach ( $data["options"] as $optionValue => $optionLabel ) {
            if ( $defaultValue !== null && $optionValue === $this->get_option( $key ) ) {
                $defaultValue = $optionValue;
            }
        }

        ob_start();
        ?>
        <tr valign="top">
            <th scope="row" class="titledesc">
                <label for="<?php echo esc_attr( $field_key ); ?>"><?php echo wp_kses_post( $data['title'] ); ?><?php echo $this->get_tooltip_html( $data ); ?></label>
            </th>
            <td class="forminp">
                <fieldset>
                    <legend class="screen-reader-text"><span><?php echo wp_kses_post( $data['title'] ); ?></span></legend>
                    <?php
                    foreach ( $data["options"] as $optionValue => $optionLabel ) {
                        ?>
                        <input id="<?php echo esc_attr( $field_key ); ?>" type="radio" name="<?php echo esc_attr( $field_key ); ?>" value="<?php echo $optionValue ?>" <?php checked( $optionValue,
                            $this->get_option( $key ), true ); ?>> <?php echo $optionLabel ?>
                        <?php
                    }
                    echo $this->get_description_html( $data );
                    ?>
                </fieldset>
            </td>
        </tr>
        <?php

        return ob_get_clean();
    }

    public
    function generate_checkbox_html(
        $key,
        $data
    ) {
        $field_key = $this->get_field_key( $key );
        $defaults  = array(
            'title'             => '',
            'label'             => '',
            'disabled'          => false,
            'class'             => '',
            'css'               => '',
            'type'              => 'text',
            'desc_tip'          => false,
            'description'       => '',
            'custom_attributes' => array(),
        );

        $data = wp_parse_args( $data, $defaults );

        if ( ! $data['label'] ) {
            $data['label'] = $data['title'];
        }

        ob_start();
        ?>
        <tr valign="top">
            <th>
                <fieldset>
                    <legend class="screen-reader-text"><span><?php echo wp_kses_post( $data['title'] ); ?></span></legend>
                    <label for="<?php echo esc_attr( $field_key ); ?>" style="white-space: nowrap">
                        <input <?php disabled( $data['disabled'], true ); ?> class="<?php echo esc_attr( $data['class'] ); ?>" type="checkbox" name="<?php echo esc_attr( $field_key ); ?>"
                                                                             id="<?php echo esc_attr( $field_key ); ?>" style="<?php echo esc_attr( $data['css'] ); ?>"
                                                                             value="1" <?php checked( $this->get_option( $key ),
                            'yes' ); ?> <?php echo $this->get_custom_attribute_html( $data ); ?> /> <?php echo wp_kses_post( $data['label'] ); ?>
                        <?php echo $this->get_tooltip_html( $data ); ?>
                    </label>
                    <?php echo $this->get_description_html( $data ); ?>
                </fieldset>
            </th>
            <td></td>
        </tr>
        <?php

        return ob_get_clean();
    }

    public function add_void_transaction_button( $order ) {
        if ( strtolower( $order->get_data()['status'] ) === strtolower( VERIFONE_TRANSACTION_PENDING ) ) {
            $label = __( 'Cancel transaction', 'verifone-hosted' );
            $slug  = 'cancel-transaction';
            $url   = get_site_url() . '/?wc-api=cancel_payment&order_id=' . $order->get_id();
            include( plugin_dir_path( __FILE__ ) . '../templates/html-void-transaction.php' );
        }
    }

    public function manual_cancel_payment() {
        ob_start();
        $params   = $_POST;
        $order_id = $params['order_id'];
        $this->cancel_payment( $order_id );
        $order = wc_get_order( $order_id );
        $note  = 'Payment transaction cancelled.';
        $order->update_status( 'cancelled', $note );
        wp_send_json( [ 'result' => 'success', 'message' => $note ], 200 );
    }

    public function hide_order_refund_button( $order ) {
        if ( ! isset( $order->get_data()['id'] ) ) {
            return;
        }
        if ( ! isset( $order->get_data()['date_paid'] ) ) {
            ?>
            <script>
                jQuery(function () {
                    jQuery('.refund-items').hide();
                    jQuery('.order_actions option[value=send_email_customer_refunded_order]').remove();
                    jQuery('#order_status option[value=wc-refunded]').remove();
                });
            </script>
            <?php
        }
    }

    private function logOrderComment( $status, $order ) {
        $note = __( "Notification event: " . $status );
        $order->add_order_note( $note, false, false );
    }

    public function stringStartsWith( $string, $needle ) {
        $len = strlen( $needle );

        return ( substr( $string, 0, $len ) === $needle );
    }


}
